"""
Custom hook to customize path for libtreelite.so
"""


def get_custom_libpath():
    """
    Get custom path for libtreelite.so.
    If valid, must return a directory containing libtreelite.so.
    """
    return None
